@@ -8,24 +8,11 @@
 --
 module Network.TLS.Core
 	(
-	-- * Context configuration
-	  TLSParams(..)
-	, TLSLogging(..)
-	, TLSCertificateUsage(..)
-	, TLSCertificateRejectReason(..)
-	, defaultLogging
-	, defaultParams
-
-	-- * Context object
-	, TLSCtx
-	, ctxConnection
-	, ctxEOF
-
 	-- * Internal packet sending and receiving
-	, sendPacket
+	  sendPacket
 	, recvPacket
 
-	-- * Creating a context
+	-- * Creating a client or server context
 	, client
 	, clientWith
 	, server
@@ -40,175 +27,34 @@ module Network.TLS.Core
 	, recvData
 	) where
 
+import Network.TLS.Context
 import Network.TLS.Struct
 import Network.TLS.Record
 import Network.TLS.Cipher
 import Network.TLS.Compression
-import Network.TLS.Crypto
 import Network.TLS.Packet
 import Network.TLS.State
 import Network.TLS.Sending
 import Network.TLS.Receiving
+import Network.TLS.Measurement
+import Network.TLS.Wire (encodeWord16)
 import Data.Maybe
-import Data.Certificate.X509
-import Data.List (intersect, intercalate, find)
+import Data.List (intersect, find)
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Lazy as L
 
 import Crypto.Random
 import Control.Applicative ((<$>))
-import Control.Concurrent.MVar
 import Control.Monad.State
-import Control.Exception (throwIO, Exception(), onException, fromException, catch)
-import Data.IORef
-import System.IO (Handle, hSetBuffering, BufferMode(..), hFlush)
+import Control.Exception (throwIO, Exception(), fromException, catch)
+import System.IO (Handle)
 import System.IO.Error (mkIOError, eofErrorType)
 import Prelude hiding (catch)
 
-data TLSLogging = TLSLogging
-	{ loggingPacketSent :: String -> IO ()
-	, loggingPacketRecv :: String -> IO ()
-	, loggingIOSent     :: Bytes -> IO ()
-	, loggingIORecv     :: Header -> Bytes -> IO ()
-	}
-
--- | Certificate and Chain rejection reason
-data TLSCertificateRejectReason =
-	  CertificateRejectExpired
-	| CertificateRejectRevoked
-	| CertificateRejectUnknownCA
-	| CertificateRejectOther String
-	deriving (Show,Eq)
-
--- | Certificate Usage callback possible returns values.
-data TLSCertificateUsage =
-	  CertificateUsageAccept                            -- ^ usage of certificate accepted
-	| CertificateUsageReject TLSCertificateRejectReason -- ^ usage of certificate rejected
-	deriving (Show,Eq)
-
-data TLSParams = TLSParams
-	{ pConnectVersion    :: Version             -- ^ version to use on client connection.
-	, pAllowedVersions   :: [Version]           -- ^ allowed versions that we can use.
-	, pCiphers           :: [Cipher]            -- ^ all ciphers supported ordered by priority.
-	, pCompressions      :: [Compression]       -- ^ all compression supported ordered by priority.
-	, pWantClientCert    :: Bool                -- ^ request a certificate from client.
-	                                            -- use by server only.
-	, pUseSecureRenegotiation :: Bool           -- notify that we want to use secure renegotation
-	, pCertificates      :: [(X509, Maybe PrivateKey)] -- ^ the cert chain for this context with the associated keys if any.
-	, pLogging           :: TLSLogging          -- ^ callback for logging
-	, onCertificatesRecv :: [X509] -> IO TLSCertificateUsage -- ^ callback to verify received cert chain.
-	}
-
-defaultLogging :: TLSLogging
-defaultLogging = TLSLogging
-	{ loggingPacketSent = (\_ -> return ())
-	, loggingPacketRecv = (\_ -> return ())
-	, loggingIOSent     = (\_ -> return ())
-	, loggingIORecv     = (\_ _ -> return ())
-	}
-
-defaultParams :: TLSParams
-defaultParams = TLSParams
-	{ pConnectVersion         = TLS10
-	, pAllowedVersions        = [TLS10,TLS11,TLS12]
-	, pCiphers                = []
-	, pCompressions           = [nullCompression]
-	, pWantClientCert         = False
-	, pUseSecureRenegotiation = True
-	, pCertificates           = []
-	, pLogging                = defaultLogging
-	, onCertificatesRecv      = (\_ -> return CertificateUsageAccept)
-	}
-
-instance Show TLSParams where
-	show p = "TLSParams { " ++ (intercalate "," $ map (\(k,v) -> k ++ "=" ++ v)
-		[ ("connectVersion", show $ pConnectVersion p)
-		, ("allowedVersions", show $ pAllowedVersions p)
-		, ("ciphers", show $ pCiphers p)
-		, ("compressions", show $ pCompressions p)
-		, ("want-client-cert", show $ pWantClientCert p)
-		, ("certificates", show $ length $ pCertificates p)
-		]) ++ " }"
-
--- | A TLS Context is a handle augmented by tls specific state and parameters
-data TLSCtx a = TLSCtx
-	{ ctxConnection      :: a             -- ^ return the connection object associated with this context
-	, ctxParams          :: TLSParams
-	, ctxState           :: MVar TLSState
-	, ctxEOF_            :: IORef Bool    -- ^ is the handle has EOFed or not.
-	, ctxConnectionFlush :: IO ()
-	, ctxConnectionSend  :: Bytes -> IO ()
-	, ctxConnectionRecv  :: Int -> IO Bytes
-	}
-
-connectionFlush :: TLSCtx c -> IO ()
-connectionFlush c = (ctxConnectionFlush c)
-
-connectionSend :: TLSCtx c -> Bytes -> IO ()
-connectionSend c b = (ctxConnectionSend c) b
-
-connectionRecv :: TLSCtx c -> Int -> IO Bytes
-connectionRecv c sz = (ctxConnectionRecv c) sz
-
-ctxEOF :: MonadIO m => TLSCtx a -> m Bool
-ctxEOF ctx = liftIO (readIORef $ ctxEOF_ ctx)
-
-throwCore :: (MonadIO m, Exception e) => e -> m a
-throwCore = liftIO . throwIO
-
-newCtxWith :: c -> IO () -> (Bytes -> IO ()) -> (Int -> IO Bytes) -> TLSParams -> TLSState -> IO (TLSCtx c)
-newCtxWith c flushF sendF recvF params st = do
-	stvar <- newMVar st
-	eof   <- newIORef False
-	return $ TLSCtx
-		{ ctxConnection = c
-		, ctxParams     = params
-		, ctxState      = stvar
-		, ctxEOF_       = eof
-		, ctxConnectionFlush = flushF
-		, ctxConnectionSend  = sendF
-		, ctxConnectionRecv  = recvF
-		}
-
-newCtx :: Handle -> TLSParams -> TLSState -> IO (TLSCtx Handle)
-newCtx handle params st = do
-	hSetBuffering handle NoBuffering
-	newCtxWith handle (hFlush handle) (B.hPut handle) (B.hGet handle) params st
-
-ctxLogging :: TLSCtx a -> TLSLogging
-ctxLogging = pLogging . ctxParams
-
-usingState :: MonadIO m => TLSCtx c -> TLSSt a -> m (Either TLSError a)
-usingState ctx f = liftIO (takeMVar mvar) >>= \st -> liftIO $ onException (execAndStore st) (putMVar mvar st)
-	where
-		mvar = ctxState ctx
-		execAndStore st = do
-			let (a, newst) = runTLSState f st
-			putMVar mvar newst
-			return a
-
-usingState_ :: MonadIO m => TLSCtx c -> TLSSt a -> m a
-usingState_ ctx f = do
-	ret <- usingState ctx f
-	case ret of
-		Left err -> throwCore err
-		Right r  -> return r
-
-getStateRNG :: MonadIO m => TLSCtx c -> Int -> m Bytes
-getStateRNG ctx n = usingState_ ctx (genTLSRandom n)
-
-whileStatus :: MonadIO m => TLSCtx c -> (TLSStatus -> Bool) -> m a -> m ()
-whileStatus ctx p a = do
-	b <- usingState_ ctx (p . stStatus <$> get)
-	when b (a >> whileStatus ctx p a)
-
 errorToAlert :: TLSError -> Packet
 errorToAlert (Error_Protocol (_, _, ad)) = Alert [(AlertLevel_Fatal, ad)]
 errorToAlert _                           = Alert [(AlertLevel_Fatal, InternalError)]
 
-setEOF :: MonadIO m => TLSCtx c -> m ()
-setEOF ctx = liftIO $ writeIORef (ctxEOF_ ctx) True
-
 readExact :: MonadIO m => TLSCtx c -> Int -> m Bytes
 readExact ctx sz = do
 	hdrbs <- liftIO $ connectionRecv ctx sz
@@ -219,33 +65,80 @@ readExact ctx sz = do
 			else throwCore (Error_Packet ("partial packet: expecting " ++ show sz ++ " bytes, got: " ++ (show $B.length hdrbs)))
 	return hdrbs
 
+recvRecord :: MonadIO m => TLSCtx c -> m (Either TLSError (Record Plaintext))
+recvRecord ctx = readExact ctx 5 >>= either (return . Left) recvLength . decodeHeader
+	where recvLength header@(Header _ _ readlen)
+		| readlen > 16384 + 2048 = return $ Left $ Error_Protocol ("record exceeding maximum size", True, RecordOverflow)
+		| otherwise              = do
+			content <- readExact ctx (fromIntegral readlen)
+			liftIO $ (loggingIORecv $ ctxLogging ctx) header content
+			usingState ctx $ disengageRecord $ rawToRecord header (fragmentCiphertext content)
+
+
 -- | receive one packet from the context that contains 1 or
 -- many messages (many only in case of handshake). if will returns a
 -- TLSError if the packet is unexpected or malformed
 recvPacket :: MonadIO m => TLSCtx c -> m (Either TLSError Packet)
 recvPacket ctx = do
-	hdrbs <- readExact ctx 5
-	case decodeHeader hdrbs of
-		Left err                          -> return $ Left err
-		Right header@(Header _ _ readlen) ->
-			if readlen > (16384 + 2048)
-				then return $ Left $ Error_Protocol ("record exceeding maximum size",True, RecordOverflow)
-				else recvLength header readlen
-	where recvLength header readlen = do
-		content <- readExact ctx (fromIntegral readlen)
-		liftIO $ (loggingIORecv $ ctxLogging ctx) header content
-		pkt <- usingState ctx $ readPacket $ rawToRecord header (fragmentCiphertext content)
-		case pkt of
-			Right p -> liftIO $ (loggingPacketRecv $ ctxLogging ctx) $ show p
-			_       -> return ()
-		return pkt
+	erecord <- recvRecord ctx
+	case erecord of
+		Left err     -> return $ Left err
+		Right record -> do
+			pkt <- usingState ctx $ processPacket record
+			case pkt of
+				Right p -> liftIO $ (loggingPacketRecv $ ctxLogging ctx) $ show p
+				_       -> return ()
+			return pkt
 
-recvPacketSuccess :: MonadIO m => TLSCtx c -> m ()
-recvPacketSuccess ctx = do
-	pkt <- recvPacket ctx
-	case pkt of
-		Left err -> throwCore err
-		Right _  -> return ()
+recvPacketHandshake :: MonadIO m => TLSCtx c -> m [Handshake]
+recvPacketHandshake ctx = do
+	pkts <- recvPacket ctx
+	case pkts of
+		Right (Handshake l) -> return l
+		Right x             -> fail ("unexpected type received. expecting handshake and got: " ++ show x)
+		Left err            -> throwCore err
+
+data RecvState m =
+	  RecvStateNext (Packet -> m (RecvState m))
+	| RecvStateHandshake (Handshake -> m (RecvState m))
+	| RecvStateDone
+
+runRecvState :: MonadIO m => TLSCtx a -> RecvState m -> m ()
+runRecvState _   (RecvStateDone)   = return ()
+runRecvState ctx (RecvStateNext f) = recvPacket ctx >>= either throwCore f >>= runRecvState ctx
+runRecvState ctx iniState          = recvPacketHandshake ctx >>= loop iniState >>= runRecvState ctx
+	where
+		loop :: MonadIO m => RecvState m -> [Handshake] -> m (RecvState m)
+		loop recvState []                  = return recvState
+		loop (RecvStateHandshake f) (x:xs) = do
+			nstate <- f x
+			usingState_ ctx $ processHandshake x
+			loop nstate xs
+		loop _                         _   = unexpected "spurious handshake" Nothing
+
+sendChangeCipherAndFinish :: MonadIO m => TLSCtx c -> Bool -> m ()
+sendChangeCipherAndFinish ctx isClient = do
+	sendPacket ctx ChangeCipherSpec
+	liftIO $ connectionFlush ctx
+	cf <- usingState_ ctx $ getHandshakeDigest isClient
+	sendPacket ctx (Handshake [Finished cf])
+	liftIO $ connectionFlush ctx
+
+recvChangeCipherAndFinish :: MonadIO m => TLSCtx c -> m ()
+recvChangeCipherAndFinish ctx = runRecvState ctx (RecvStateNext expectChangeCipher)
+	where
+		expectChangeCipher ChangeCipherSpec = return $ RecvStateHandshake expectFinish
+		expectChangeCipher p                = unexpected (show p) (Just "change cipher")
+		expectFinish (Finished _) = return RecvStateDone
+		expectFinish p            = unexpected (show p) (Just "Handshake Finished")
+
+unexpected :: MonadIO m => String -> Maybe [Char] -> m a
+unexpected msg expected = throwCore $ Error_Packet_unexpected msg (maybe "" (" expected: " ++) expected)
+
+newSession :: MonadIO m => TLSCtx c -> m Session
+newSession ctx
+	| pUseSession $ ctxParams ctx = getStateRNG ctx 32 >>= return . Session . Just
+	| otherwise                   = return $ Session Nothing
 
 -- | Send one packet to the context
 sendPacket :: MonadIO m => TLSCtx c -> Packet -> m ()
@@ -287,45 +180,36 @@ server params rng handle = liftIO $ newCtx handle params st
 bye :: MonadIO m => TLSCtx c -> m ()
 bye ctx = sendPacket ctx $ Alert [(AlertLevel_Warning, CloseNotify)]
 
+-- | when a new handshake is done, wrap up & clean up.
+handshakeTerminate :: MonadIO m => TLSCtx c -> m ()
+handshakeTerminate ctx = do
+	session <- usingState_ ctx getSession
+	-- only callback the session established if we have a session
+	case session of
+		Session (Just sessionId) -> do
+			sessionData <- usingState_ ctx getSessionData
+			liftIO $ (onSessionEstablished $ ctxParams ctx) sessionId (fromJust sessionData)
+		_ -> return ()
+	-- forget all handshake data now and reset bytes counters.
+	usingState_ ctx endHandshake
+	updateMeasure ctx resetBytesCounters
+	return ()
+
 -- client part of handshake. send a bunch of handshake of client
 -- values intertwined with response from the server.
 handshakeClient :: MonadIO m => TLSCtx c -> m ()
 handshakeClient ctx = do
-	-- Send ClientHello
-	crand <- getStateRNG ctx 32 >>= return . ClientRandom
-	extensions <- getExtensions
-	usingState_ ctx (startHandshakeClient ver crand)
-	sendPacket ctx $ Handshake
-		[ ClientHello ver crand (Session Nothing) (map cipherID ciphers)
-		              (map compressionID compressions) extensions
-		]
-
-	-- Receive Server information until ServerHelloDone
-	whileStatus ctx (/= (StatusHandshake HsStatusServerHelloDone)) $ do
-		pkts <- recvPacket ctx
-		case pkts of
-			Left err -> throwCore err
-			Right l  -> processServerInfo l
-
-	-- Send Certificate if requested. XXX disabled for now.
-	certRequested <- return False
-	when certRequested (sendPacket ctx $ Handshake [Certificates clientCerts])
-
-	sendClientKeyXchg
-
-	{- maybe send certificateVerify -}
-	{- FIXME not implemented yet -}
-
-	sendPacket ctx ChangeCipherSpec
-	liftIO $ connectionFlush ctx
-
-	-- Send Finished
-	cf <- usingState_ ctx $ getHandshakeDigest True
-	sendPacket ctx (Handshake [Finished cf])
-
-	-- receive changeCipherSpec & Finished
-	recvPacketSuccess ctx >> recvPacketSuccess ctx >> return ()
-
+	updateMeasure ctx incrementNbHandshakes
+	sendClientHello
+	recvServerHello
+	sessionResuming <- usingState_ ctx isSessionResuming
+	if sessionResuming
+		then sendChangeCipherAndFinish ctx True
+		else do
+			sendCertificate >> sendClientKeyXchg >> sendCertificateVerify
+			sendChangeCipherAndFinish ctx True
+			recvChangeCipherAndFinish ctx
+	handshakeTerminate ctx
 	where
 		params       = ctxParams ctx
 		ver          = pConnectVersion params
@@ -338,10 +222,35 @@ handshakeClient ctx = do
 			then usingState_ ctx (getVerifiedData True) >>= \vd -> return [ (0xff01, encodeExtSecureRenegotiation vd Nothing) ]
 			else return []
 
-		processServerInfo (Handshake hss) = mapM_ processHandshake hss
-		processServerInfo _               = return ()
+		sendClientHello = do
+			crand <- getStateRNG ctx 32 >>= return . ClientRandom
+			let clientSession = Session . maybe Nothing (Just . fst) $ sessionResumeWith params
+			extensions <- getExtensions
+			usingState_ ctx (startHandshakeClient ver crand)
+			sendPacket ctx $ Handshake
+				[ ClientHello ver crand clientSession (map cipherID ciphers)
+					      (map compressionID compressions) extensions
+				]
 
-		processHandshake (ServerHello rver _ _ cipher _ _) = do
+		expectChangeCipher ChangeCipherSpec = return $ RecvStateHandshake expectFinish
+		expectChangeCipher p                = unexpected (show p) (Just "change cipher")
+		expectFinish (Finished _) = return RecvStateDone
+		expectFinish p            = unexpected (show p) (Just "Handshake Finished")
+
+		sendCertificate = do
+			-- Send Certificate if requested. XXX disabled for now.
+			certRequested <- return False
+			when certRequested (sendPacket ctx $ Handshake [Certificates clientCerts])
+
+		sendCertificateVerify =
+			{- maybe send certificateVerify -}
+			{- FIXME not implemented yet -}
+			return ()
+
+		recvServerHello = runRecvState ctx (RecvStateHandshake onServerHello)
+
+		onServerHello :: MonadIO m => Handshake -> m (RecvState m)
+		onServerHello sh@(ServerHello rver _ serverSession cipher _ _) = do
 			when (rver == SSL2) $ throwCore $ Error_Protocol ("ssl2 is not supported", True, ProtocolVersion)
 			case find ((==) rver) allowedvers of
 				Nothing -> throwCore $ Error_Protocol ("version " ++ show ver ++ "is not supported", True, ProtocolVersion)
@@ -350,21 +259,55 @@ handshakeClient ctx = do
 				Nothing -> throwCore $ Error_Protocol ("no cipher in common with the server", True, HandshakeFailure)
 				Just c  -> usingState_ ctx $ setCipher c
 
-		processHandshake (Certificates certs) = do
+			let resumingSession = case sessionResumeWith params of
+				Just (sessionId, sessionData) -> if serverSession == Session (Just sessionId) then Just sessionData else Nothing
+				Nothing                       -> Nothing
+			usingState_ ctx $ setSession serverSession (isJust resumingSession)
+			usingState_ ctx $ processServerHello sh
+			case resumingSession of
+				Nothing          -> return $ RecvStateHandshake processCertificate
+				Just sessionData -> do
+					usingState_ ctx (setMasterSecret $ sessionSecret sessionData)
+					return $ RecvStateNext expectChangeCipher
+		onServerHello p = unexpected (show p) (Just "server hello")
+
+		processCertificate :: MonadIO m => Handshake -> m (RecvState m)
+		processCertificate (Certificates certs) = do
 			let cb = onCertificatesRecv $ params
 			usage <- liftIO $ cb certs
 			case usage of
 				CertificateUsageAccept        -> return ()
 				CertificateUsageReject reason -> certificateRejected reason
+			return $ RecvStateHandshake processServerKeyExchange
+		processCertificate p = processServerKeyExchange p
 
-		processHandshake (CertRequest _ _ _) = do
-			return ()
+		processServerKeyExchange :: MonadIO m => Handshake -> m (RecvState m)
+		processServerKeyExchange (ServerKeyXchg _) = return $ RecvStateHandshake processCertificateRequest
+		processServerKeyExchange p                 = processCertificateRequest p
+
+		processCertificateRequest (CertRequest _ _ _) = do
 			--modify (\sc -> sc { scCertRequested = True })
-		processHandshake _ = return ()
+			return $ RecvStateHandshake processServerHelloDone
+		processCertificateRequest p = processServerHelloDone p
+
+		processServerHelloDone ServerHelloDone = return RecvStateDone
+		processServerHelloDone p = unexpected (show p) (Just "server hello data")
 
 		sendClientKeyXchg = do
-			prerand <- getStateRNG ctx 46 >>= return . ClientKeyData
-			sendPacket ctx $ Handshake [ClientKeyXchg ver prerand]
+			encryptedPreMaster <- usingState_ ctx $ do
+				xver       <- stVersion <$> get
+				prerand    <- genTLSRandom 46
+				let premaster = encodePreMasterSecret xver prerand
+				setMasterSecretFromPre premaster
+
+				-- SSL3 implementation generally forget this length field since it's redundant,
+				-- however TLS10 make it clear that the length field need to be present.
+				e <- encryptRSA premaster
+				let extra = if xver < TLS10
+					then B.empty
+					else encodeWord16 $ fromIntegral $ B.length e
+				return $ extra `B.append` e
+			sendPacket ctx $ Handshake [ClientKeyXchg encryptedPreMaster]
 
 		-- on certificate reject, throw an exception with the proper protocol alert error.
 		certificateRejected CertificateRejectRevoked =
@@ -377,8 +320,14 @@ handshakeClient ctx = do
 			throwCore $ Error_Protocol ("certificate rejected: " ++ s, True, CertificateUnknown)
 
 handshakeServerWith :: MonadIO m => TLSCtx c -> Handshake -> m ()
-handshakeServerWith ctx (ClientHello ver _ _ ciphers compressions _) = do
+handshakeServerWith ctx clientHello@(ClientHello ver _ clientSession ciphers compressions _) = do
+	-- check if policy allow this new handshake to happens
+	handshakeAuthorized <- withMeasure ctx (onHandshake $ ctxParams ctx)
+	unless handshakeAuthorized (throwCore $ Error_HandshakePolicy "server: handshake denied")
+	updateMeasure ctx incrementNbHandshakes
+
 	-- Handle Client hello
+	usingState_ ctx $ processHandshake clientHello
 	when (ver == SSL2) $ throwCore $ Error_Protocol ("ssl2 is not supported", True, ProtocolVersion)
 	when (not $ elem ver (pAllowedVersions params)) $
 		throwCore $ Error_Protocol ("version " ++ show ver ++ "is not supported", True, ProtocolVersion)
@@ -392,21 +341,25 @@ handshakeServerWith ctx (ClientHello ver _ _ ciphers compressions _) = do
 		, stCompression = usedCompression
 		})
 
-	-- send Server Data until ServerHelloDone
-	handshakeSendServerData
-	liftIO $ connectionFlush ctx
+	resumeSessionData <- case clientSession of
+		(Session (Just clientSessionId)) -> liftIO $ onSessionResumption params $ clientSessionId
+		(Session Nothing)                -> return Nothing
+	case resumeSessionData of
+		Nothing -> do
+			handshakeSendServerData
+			liftIO $ connectionFlush ctx
 
-	-- Receive client info until client Finished.
-	whileStatus ctx (/= (StatusHandshake HsStatusClientFinished)) (recvPacketSuccess ctx)
-
-	sendPacket ctx ChangeCipherSpec
-
-	-- Send Finish
-	cf <- usingState_ ctx $ getHandshakeDigest False
-	sendPacket ctx (Handshake [Finished cf])
-
-	liftIO $ connectionFlush ctx
-	return ()
+			-- Receive client info until client Finished.
+			recvClientData
+			sendChangeCipherAndFinish ctx False
+		Just sessionData -> do
+			usingState_ ctx (setSession clientSession True)
+			serverhello <- makeServerHello clientSession
+			sendPacket ctx $ Handshake [serverhello]
+			usingState_ ctx $ setMasterSecret $ sessionSecret sessionData
+			sendChangeCipherAndFinish ctx False
+			recvChangeCipherAndFinish ctx
+	handshakeTerminate ctx
 	where
 		params             = ctxParams ctx
 		commonCiphers      = intersect ciphers (map cipherID $ pCiphers params)
@@ -417,17 +370,33 @@ handshakeServerWith ctx (ClientHello ver _ _ ciphers compressions _) = do
 		privKeys           = map snd $ pCertificates params
 		needKeyXchg        = cipherExchangeNeedMoreData $ cipherKeyExchange usedCipher
 
-		handshakeSendServerData = do
-			srand <- getStateRNG ctx 32 >>= return . ServerRandom
+		---
+		recvClientData = runRecvState ctx (RecvStateHandshake $ processClientCertificate)
 
+		processClientCertificate (Certificates _) = return $ RecvStateHandshake processClientKeyExchange
+		processClientCertificate p = processClientKeyExchange p
+
+		processClientKeyExchange (ClientKeyXchg _) = return $ RecvStateNext processCertificateVerify
+		processClientKeyExchange p                 = unexpected (show p) (Just "client key exchange")
+
+		processCertificateVerify (Handshake [CertVerify _]) = return $ RecvStateNext expectChangeCipher
+		processCertificateVerify p = expectChangeCipher p
+
+		expectChangeCipher ChangeCipherSpec = return $ RecvStateHandshake expectFinish
+		expectChangeCipher p                = unexpected (show p) (Just "change cipher")
+
+		expectFinish (Finished _) = return RecvStateDone
+		expectFinish p            = unexpected (show p) (Just "Handshake Finished")
+		---
+
+		makeServerHello session = do
+			srand <- getStateRNG ctx 32 >>= return . ServerRandom
 			case privKeys of
 				(Just privkey : _) -> usingState_ ctx $ setPrivateKey privkey
 				_                  -> return () -- return a sensible error
 
 			-- in TLS12, we need to check as well the certificates we are sending if they have in the extension
 			-- the necessary bits set.
-
-			-- send ServerHello & Certificate & ServerKeyXchg & CertReq
 			secReneg   <- usingState_ ctx getSecureRenegotiation
 			extensions <- if secReneg
 				then do
@@ -438,11 +407,15 @@ handshakeServerWith ctx (ClientHello ver _ _ ciphers compressions _) = do
 					return [ (0xff01, vf) ]
 				else return []
 			usingState_ ctx (setVersion ver >> setServerRandom srand)
-			sendPacket ctx $ Handshake
-				[ ServerHello ver srand (Session Nothing) (cipherID usedCipher)
-				                        (compressionID usedCompression) extensions
-				, Certificates srvCerts
-				]
+			return $ ServerHello ver srand session (cipherID usedCipher)
+			                               (compressionID usedCompression) extensions
+
+		handshakeSendServerData = do
+			serverSession <- newSession ctx
+			usingState_ ctx (setSession serverSession False)
+			serverhello   <- makeServerHello serverSession
+			-- send ServerHello & Certificate & ServerKeyXchg & CertReq
+			sendPacket ctx $ Handshake [ serverhello, Certificates srvCerts ]
 			when needKeyXchg $ do
 				let skg = SKX_RSA Nothing
 				sendPacket ctx (Handshake [ServerKeyXchg skg])
@@ -456,13 +429,13 @@ handshakeServerWith ctx (ClientHello ver _ _ ciphers compressions _) = do
 
 handshakeServerWith _ _ = fail "unexpected handshake type received. expecting client hello"
 
--- after receiving a client hello, we need to redo a handshake -}
+-- after receiving a client hello, we need to redo a handshake
 handshakeServer :: MonadIO m => TLSCtx c -> m ()
 handshakeServer ctx = do
-	pkts <- recvPacket ctx
-	case pkts of
-		Right (Handshake [hs]) -> handshakeServerWith ctx hs
-		x                      -> fail ("unexpected type received. expecting handshake ++ " ++ show x)
+	hss <- recvPacketHandshake ctx
+	case hss of
+		[ch] -> handshakeServerWith ctx ch
+		_    -> fail ("unexpected handshake received, excepting client hello and received " ++ show hss)
 
 -- | Handshake for a new TLS connection
 -- This is to be called at the beginning of a connection, and during renegociation
@@ -474,7 +447,7 @@ handshake ctx = do
 		handleException f = catch (f >> return True) (\e -> handler e >> return False)
 		handler e = case fromException e of
 			Just err -> sendPacket ctx (errorToAlert err)
-			Nothing  -> sendPacket ctx (errorToAlert $ Error_Misc "")
+			Nothing  -> sendPacket ctx (errorToAlert $ Error_Misc $ show e)
 
 -- | sendData sends a bunch of data.
 -- It will automatically chunk data to acceptable packet size
