@@ -28,11 +28,12 @@ module Network.TLS.Struct
 	, Header(..)
 	, ServerRandom(..)
 	, ClientRandom(..)
-	, ClientKeyData(..)
 	, serverRandom
 	, clientRandom
 	, FinishedData
+	, SessionID
 	, Session(..)
+	, SessionData(..)
 	, AlertLevel(..)
 	, AlertDescription(..)
 	, HandshakeType(..)
@@ -57,8 +58,6 @@ type Bytes = ByteString
 -- | Versions known to TLS
 --
 -- SSL2 is just defined, but this version is and will not be supported.
---
--- TLS12 is not yet supported
 data Version = SSL2 | SSL3 | TLS10 | TLS11 | TLS12 deriving (Show, Eq, Ord)
 
 data ConnectionEnd = ConnectionServer | ConnectionClient
@@ -112,6 +111,7 @@ data TLSError =
 	  Error_Misc String        -- ^ mainly for instance of Error
 	| Error_Protocol (String, Bool, AlertDescription)
 	| Error_Certificate String
+	| Error_HandshakePolicy String -- ^ handshake policy failed.
 	| Error_Random String
 	| Error_EOF
 	| Error_Packet String
@@ -140,8 +140,15 @@ data Header = Header ProtocolType Version Word16 deriving (Show,Eq)
 
 newtype ServerRandom = ServerRandom Bytes deriving (Show, Eq)
 newtype ClientRandom = ClientRandom Bytes deriving (Show, Eq)
-newtype ClientKeyData = ClientKeyData Bytes deriving (Show, Eq)
-newtype Session = Session (Maybe Bytes) deriving (Show, Eq)
+type SessionID = Bytes
+newtype Session = Session (Maybe SessionID) deriving (Show, Eq)
+
+data SessionData = SessionData
+	{ sessionVersion :: Version
+	, sessionCipher  :: CipherID
+	, sessionSecret  :: Bytes
+	}
+
 type CipherID = Word16
 type CompressionID = Word8
 type FinishedData = Bytes
@@ -227,7 +234,7 @@ data Handshake =
 	| Certificates [X509]
 	| HelloRequest
 	| ServerHelloDone
-	| ClientKeyXchg Version ClientKeyData
+	| ClientKeyXchg Bytes
 	| ServerKeyXchg ServerKeyXchgAlgorithmData
 	| CertRequest [CertificateType] (Maybe [ (HashAlgorithm, SignatureAlgorithm) ]) [Word8]
 	| CertVerify [Word8]
@@ -246,7 +253,7 @@ typeOfHandshake (ServerHello _ _ _ _ _ _) = HandshakeType_ServerHello
 typeOfHandshake (Certificates _)          = HandshakeType_Certificate
 typeOfHandshake (HelloRequest)            = HandshakeType_HelloRequest
 typeOfHandshake (ServerHelloDone)         = HandshakeType_ServerHelloDone
-typeOfHandshake (ClientKeyXchg _ _)       = HandshakeType_ClientKeyXchg
+typeOfHandshake (ClientKeyXchg _)         = HandshakeType_ClientKeyXchg
 typeOfHandshake (ServerKeyXchg _)         = HandshakeType_ServerKeyXchg
 typeOfHandshake (CertRequest _ _ _)       = HandshakeType_CertRequest
 typeOfHandshake (CertVerify _)            = HandshakeType_CertVerify
